module HaskellIntro where

import Data.Char

-- TYPES & FUNCTIONS

-- 1. Types!
--

-- * Everything* has a type.  A type is a guarantee about the behavior

-- of a program, which can be checked before actually running the
-- program.

-- We can define things in Haskell as follows:

n :: Integer -- "n has type Integer" (type signature)
n = 5 -- "n is defined to be 5".

-- In Haskell, = really means =.  The above is a
-- _definition_, not an _assignment_.

-- This would be an error:
-- n = 10

z :: Integer
z = 3 * n + 5

-- Q: How does Haskell evaluate z?
-- A: Via rewriting and substitution!

{-
     z
 --> 3 * n + 5
 --> 3 * 5 + 5
 --> 15 + 5
 --> 20
-}

c :: Char
c = 't' -- single quotes around any Unicode character

b :: Bool
b = False -- or True

d :: Double -- double-precision floating point numbers
d = 3.92 + 1.0e-3

m :: Int -- machine-word-sized integer
m = 23

s :: String
s = "Hello, CSCI 365!" -- Strings use double quotes.

------------------------------------------------------------
-- Functions

-- What is a function?
-- - Code that takes some inputs, does computation,
--   returns a result
-- - Process with input + output

{- Is this a function?

def foo(x):
  print(f'x = {x}')

In Python, this is called a function, but it's not a function in the
mathematical sense, because it has some extra "side effects" besides just computing a result.

In Haskell, we *only* have functions in the mathematical sense --- Haskell functions take inputs, and produce outputs, and nothing else.

-}

-- We define the variable 'square' to have type
-- "Integer -> Integer", that is, square is
-- a function that takes an Integer as input
-- and returns an Integer as output.
square :: Integer -> Integer
square x = x * x

-- To apply a function to an input, just write them
-- with a space in between.

area :: Integer
area = square 4 + square 7

-- f n-1 = (f n) - 1
-- f (n-1)

{-

  area
  --> square 4 + square 7
  --> (4 * 4) + square 7
  --> 16 + square 7
  --> 16 + (7 * 7)
  --> 16 + 49
  --> 65
-}

factorial :: Integer -> Integer
factorial 0 = 1
factorial n = n * factorial (n - 1)

{-

  factorial 3
  --> 3 * factorial (3-1)
  --> 3 * factorial 2
  --> 3 * (2 * factorial (2-1))
  --> 3 * (2 * factorial 1)
  --> 3 * (2 * (1 * factorial (1-1)))
  --> 3 * (2 * (1 * factorial 0))
  --> 3 * (2 * (1 * 1))
  --> ...
  --> 6
-}

myNot :: Bool -> Bool
myNot False = True
myNot True = False

myNot2 :: Bool -> Bool
myNot2 False = True
myNot2 _ = False -- _ is a "wildcard pattern" --- don't care what the input is

myXor :: Bool -> Bool -> Bool -- multi-argument function types have all inputs separated by ->
myXor True False = True -- To apply a multi-argument function, just separate arguments with space
myXor False True = True
myXor _ _ = False

-- Haskell has built-in  && ,  ||, not

myXor2 :: Bool -> Bool -> Bool
myXor2 x y = (x && not y) || (not x && y)

-- if expressions:
--
-- e ::= ..... | if e1 then e2 else e3
--
-- if True then e2 else e3 ---> e2
-- if False then e2 else e3 ---> e3

-- mod :: Integer -> Integer -> Integer
-- Note / is division on e.g. Double only
-- For integer division, we can use the 'div' function
--
-- div :: Integer -> Integer -> Integer

-- hailstone function: if even -> n/2, else 3n+1
hailstone :: Integer -> Integer
hailstone n = if mod n 2 == 0 then div n 2 else 3 * n + 1

-- You can turn function names into infix operators
-- by surrounding them in backticks
--
-- e.g.  mod n 2    vs    n `mod` 2

hailstone2 :: Integer -> Integer
hailstone2 n = if n `mod` 2 == 0 then n `div` 2 else 3 * n + 1

-- There is a built-in library function
-- even :: Integer -> Bool

hailstone3 :: Integer -> Integer
hailstone3 n = if even n then n `div` 2 else 3 * n + 1

-- Instead of using 'if' we can also use *guards*:
-- we pick the first guard that is True.

hailstone4 :: Integer -> Integer
hailstone4 n
  | even n = n `div` 2
  | otherwise = 3 * n + 1

-- < > <= >= == /=

myIsDigit :: Char -> Bool
myIsDigit c = '0' <= c && c <= '9'

caesar :: Int -> Char -> Char
caesar k c = chr ((((ord c - ord 'A') + k) `mod` 26) + ord 'A')

------------------------------------------------------------

fib :: Integer -> Integer
fib 0 = 0
fib 1 = 1
fib n = fib (n - 1) + fib (n - 2)

-- (F_{n-1}, F_n) ---> (F_n, F_{n+1})

fibStep :: (Integer, Integer) -> (Integer, Integer)
fibStep (f1, f2) = (f2, f1 + f2)

-- Could also write it this way, which is ugly:
-- fibStepUgly p = (snd p, fst p + snd p)

fib2 :: Integer -> Integer
fib2 n = fst (fib2Helper n (0, 1))

-- fib2 n = fn
--   where
--     (fn, _) = fib2Helper n (0, 1)

-- fib2Helper n (fa, fb) will iterate fibStep on
-- the pair of Fibonacci numbers n times.
-- e.g.
-- fib2Helper 1 (x,y) = fibStep (x,y)
-- fib2Helper 2 (x,y) = fibStep (fibStep (x,y))
fib2Helper :: Integer -> (Integer, Integer) -> (Integer, Integer)
fib2Helper 0 fibs = fibs
fib2Helper n fibs = fib2Helper (n - 1) (fibStep fibs)

------------------------------------------------------------

swap :: (Int, Int) -> (Int, Int)
swap (x, y) = (y, x)

sortTup :: (Int, Int) -> (Int, Int)
sortTup (x, y)
  | x < y = (x, y)
  | otherwise = (y, x)

sortTup :: (Int, Int) -> (Int, Int)
sortTup (x, y) = (if x < y then id else swap) (x, y)

diff :: (Int, Int) -> Int
diff (x, y) = y - x

absDiff1 :: (Int, Int) -> Int
absDiff1 (x, y) = abs (y - x)

absDiff2 :: (Int, Int) -> Int
absDiff2 p = abs (diff p)

-- (f . g) x = f (g x)

absDiff3 :: (Int, Int) -> Int
absDiff3 p = (abs . diff) p

absDiff4 :: (Int, Int) -> Int
absDiff4 = abs . diff

absDiff5 :: (Int, Int) -> Int
absDiff5 = diff . sortTup
